package D2P::Entity::Schema;
use base qw(D2P::Entity::Module);


sub	init
{
	my	$this = shift;

	$this->SUPER::init('Schema', @_);
	$this->{type} = 'Database schema';
	$this->{header} = "";
	$this->{modules} = {
		byname	=> {},
		byid	=> {},
		ordered	=> []
	};
	$this->{extModules} = {
		byname	=> {},
		byid	=> {}
	};
}


sub	addModule
{
	my	$this	= shift;
	my	$module	= shift;

	if	($module->name() eq "")
	{
		print STDERR "WARNING : name not set for module\n";
		return 0;
	}
	elsif	($module->diaId() eq "")
	{
		print STDERR "WARNING : DIA identifier not set for module\n";
		return 0;
	}
	elsif	(	defined($this->{modules}->{byname}->{$module->name()})
		||	defined($this->{extModules}->{byname}->{$module->name()})
		)
	{
		print STDERR "WARNING : a module named '" . $module->name() . "' already exists\n";
		return 0;
	}
	elsif	(	defined($this->{modules}->{byid}->{$module->diaId()})
		||	defined($this->{extModules}->{byid}->{$module->diaId()})
		)
	{
		print STDERR "WARNING : a module with id '" . $module->diaId() . "' already exists\n";
		return 0;
	}

	my	$dest = $module->isExternal() ? $this->{extModules} : $this->{modules};
	$dest->{byid}->{$module->diaId()} = $module;
	$dest->{byname}->{$module->name()} = $module;
	foreach my $tid (keys %{$module->{tables}->{byid}})
	{
		my	$t = $module->{tables}->{byid}->{$tid};
		$this->addTable($t);
		$t->module($module);
	}
	foreach my $rid (keys %{$module->{relations}})
	{
		my	$r = $module->{relations}->{$rid};
		$this->addRelation($r);
		$r->module($module);
	}
	foreach my $cid (keys %{$module->{comments}})
	{
		my	$c = $module->{comments}->{$cid};
		$this->{comments}->{$cid} = $c;
	}
	return 1;
}


sub	addModuleCopy
{
	my	$this	= shift;
	my	$module	= shift;

	if	($module->name() eq "")
	{
		print STDERR "WARNING : name not set for module\n";
		return 0;
	}
	elsif	($module->diaId() eq "")
	{
		print STDERR "WARNING : DIA identifier not set for module\n";
		return 0;
	}
	elsif	(	defined($this->{modules}->{byname}->{$module->name()})
		||	defined($this->{extModules}->{byname}->{$module->name()})
		)
	{
		print STDERR "WARNING : a module named '" . $module->name() . "' already exists\n";
		return 0;
	}

	my	$dest	= $module->isExternal() ? $this->{extModules} : $this->{modules};
	my	$nb	= keys(%{$dest->{byname}});
	$dest->{byname}->{$module->name()} = $module;
	$dest->{byid}->{$nb . "_" . $module->diaId()} = $module;
	$module->diaId($nb . "_" . $module->diaId());
	foreach my $tid (keys %{$module->{tables}->{byid}})
	{
		my	$t = $module->{tables}->{byid}->{$tid};
		$this->{tables}->{byname}->{$t->name()} = $t;
		$this->{tables}->{byid}->{$module->name() . $tid} = $t;
		$module->{tables}->{byid}->{$module->name() . $tid} = $t;
		$t->diaId($module->name() . $tid);
		undef $module->{tables}->{byid}->{$tid};
		delete $module->{tables}->{byid}->{$tid};

		foreach my $fld (@{$t->{fields}})
		{
			$fld->{reference} = undef;
		}
	}
	foreach my $cid (keys %{$module->{comments}})
	{
		my	$c = $module->{comments}->{$cid};
		$this->{comments}->{$module->name() . $cid} = $c;
		$c->diaId($module->name() . $cid);
		$module->{comments}->{$module->name() . $cid} = $c;
		undef $module->{comments}->{$cid};
		delete $module->{comments}->{$cid};
	}
	foreach my $rid (keys %{$module->{relations}})
	{
		my	$r = $module->{relations}->{$rid};
		$this->{relations}->{$module->name() . $rid} = $r;
		defined($r->fromId())
			&&	(	defined($this->{tables}->{byid}->{$module->name() . $r->fromId()})
				||	defined($this->{comments}->{$module->name() . $r->fromId()})
				)
			&& $r->fromId($module->name() . $r->fromId());
		defined($r->toId())
			&&	(	defined($this->{tables}->{byid}->{$module->name() . $r->toId()})
				||	defined($this->{comments}->{$module->name() . $r->toId()})
				)
			&& $r->toId($module->name() . $r->toId());
		$r->diaId($module->name() . $rid);
		$module->{relations}->{$module->name() . $rid} = $r;
		undef $module->{relations}->{$rid};
		delete $module->{relations}->{$rid};
	}
	return 1;
}


sub	resolveRelations
{
	my	$this	= shift;
	my	%prio	= (
		"D2P::Entity::Relation::Comment"	=>	2,
		"D2P::Entity::Relation::Reference"	=>	1,
		"D2P::Entity::Relation::Inheritance"	=>	0
	);

	foreach my $relId (sort { $prio{ref($this->{relations}->{$a})} <=> $prio{ref($this->{relations}->{$b})} }
			keys %{$this->{relations}}
	)
	{
		my	$rel	= $this->{relations}->{$relId};

		if	($prio{ref($rel)} == 2)
		{
			# Handle comments
			my	$st	= $this->{tables}->{byid}->{$rel->fromId()};
			my	$dt	= $this->{comments}->{$rel->toId()};

			unless	(defined($st))
			{
				print STDERR "WARNING : Table " . $rel->fromId() . " not found for comment $relId\n";
				next;
			}
			unless	(defined($dt))
			{
				print STDERR "WARNING : Comment " . $rel->toId() . " not found for comment $relId\n";
				next;
			}

			$rel->setTables($st, $dt);
		}
		else
		{
			# Handle references and inheritance
			my	$st	= $this->{tables}->{byid}->{$rel->fromId()};
			my	$dt	= $this->{tables}->{byid}->{$rel->toId()};

			unless	(defined($st) && defined($dt))
			{
				print STDERR "WARNING : Tables not found for relation $relId\n";
				next;
			}

			if	(ref($rel->module()) =~ /::Schema$/)
			{
				$st->module()->addRelation($rel);
				if	($dt->module->isExternal())
				{
					$rel->module()->extRelCnt($rel->module()->extRelCnt() + 1);
				}
				else
				{
					$rel->module()->modRelCnt($rel->module()->modRelCnt() + 1);
				}
			}
			$rel->setTables($st, $dt);
		}
	}
}


sub	reorder
{
	my	$this	= shift;
	my	(@all, @ordered, %relations);

	## Find modules and their relations
	foreach my $mnm (sort keys %{$this->{modules}->{byname}})
	{
		my	$m = $this->{modules}->{byname}->{$mnm};
		push @all, $m;

		my	@rels = ();
		foreach my $r (values %{$m->{relations}})
		{
			my	$from	= $r->from();
			my	$to	= $r->to();
			next unless $from && $to && (ref($to) =~ /^D2P::Entity::Table/);
			my	$frmod	= $from->module();
			my	$tomod	= $to->module();
			next unless $frmod && ($frmod->name() eq $mnm);
			next unless $tomod && ($tomod->name() ne $mnm);
			push @rels, $tomod->name()
				unless grep { $_ eq $tomod->name() } @rels;
		}
		$relations{$mnm} = [@rels];
	}

	## Order modules
	my	$relNb = 0;
	@ordered = ();
	while	(@all)
	{
		my	@keep = ();
		foreach my $m (@all)
		{
			my	@rels = @{$relations{$m->name()}};
			if	(!@rels)
			{
				push @ordered, $m;
				next;
			}

			my	$pres = 0;
			foreach my $rn (@rels)
			{
				$pres ++ if	(	defined($this->{extModules}->{byname}->{$rn})
						||	grep { $_->name() eq $rn } @ordered
						);
			}

			if	($pres == @rels)
			{
				push @ordered, $m;
			}
			else
			{
				push @keep, $m;
			}
		}
		if	(@all > @keep)
		{
			@all = @keep;
		}
		else
		{
			die "ERROR : modules have redundant dependencies\n";
		}
	}
	@{$this->{modules}->{ordered}} = @ordered;

	foreach my $mod (@ordered)
	{
		$mod->reorder();
	}
}

sub	postprocess
{
	my	$this	= shift;

	## Resolve relations and sort modules/tables by creation order
	$this->resolveRelations();
	$this->reorder();

	## Postprocess modules
	foreach my $mod (@{$this->{modules}->{ordered}})
	{
		$mod->postprocess();
	}
}


1;
